#!/usr/bin/perl -w
use strict;

# PCR error rate - pure guesswork.  At the very least set this to 1e-15 in
# order to prevent underflows in the floating point code as it's not
# particularly rigorous.
#
# The affect of reducing this is to increase Q1 scores and consequentially
# also give more variability to U2 (SAM Q2/Q3) scores.

my $pPCR = 2e-5;  # 1 in 50,000

#-----------------------------------------------------------------------------
# SAM line generation
sub output_sam {
    local $"=",";
    my ($end, $name,$rname,$rg,$pos,$strand,$s1,$q1,$bpc, @c1) = @_;
    my $flag = 1; #paired
    my $cigar = length($s1) . "M";

    my $aux="RG:Z:$rg\tEV:Z:@c1";

#    my $hex;
#    if ($bpc == 1) {
#	$hex = join("", map { sprintf("%02X", $_) } @c1);
#    } elsif ($bpc == 2) {
#	$hex = join("", map { sprintf("%04X", $_) } @c1);
#    } else {
#	$hex = join("", map { sprintf("%08X", $_) } @c1);
#    }
#    my $aux="RG:Z:$rg\tEV:Z:$hex";

    $flag += 0x10 if (($strand eq "-") != ($end == 1));
    $flag += $end ? 0x80 : 0x40;

    print "$name\t$flag\t$rname\t$pos\t255\t$cigar\t=\t0\t0\t$s1\t$q1\t$aux\n";
}

#-----------------------------------------------------------------------------
# Arg handling
if (scalar(@ARGV) != 2 ) {
    print STDERR "Usage: likeC2sam library_file.gz ev.codes [a.likeC ...]\n";
    exit 1
}
my $rg = shift(@ARGV); #Library file
my $ev = shift(@ARGV); #EV table
open(my $fh, "gzip -cd < $rg |") || die "$rg";

my $line = 0;

# SAM header
print "\@SQ\tSN:NA19240_chr20_50000_10049999_hap1\tLN:10000000\n";
my $rg_done = 0;

#-----------------------------------------------------------------------------
# Main processing loop

my $max = 0; # Largest code seen so far.

while (<>) {
    my @F = split("\t", $_);
    my $lib_name = <$fh>;
    my $lib_seq  = <$fh>;

    $line++;

    # Skip filtered data
    next if scalar(@F) <= 4;

    # Parse library name line
    $lib_name =~ />(\S+).*Strand (.) .* (\d+)--(\d+)/;
    my ($rname,$strand,$start,$end) = ($1,$2,$3,$4);

    if (!$rg_done) {
	print "\@RG\tID:$rname\tSM:Test_data\tEV:$ev\n";
	$rg_done = 1;
    }


    if ($strand eq "-") {
	my ($rst,$ren) = ($rname =~ /_(\d+)_(\d+)_/);
	my $len = $end - $start;

	$start = ($ren-$rst) - $end + 2;
	$end   = $start + $len;
    } else {
	$start++;
    }

    # This is the correct name from illumina format data, but we seem to have
    # collisions. I expect the names have been generated randomly in
    # the simulated test set.
    my $name = "$F[0]_$F[1]_$F[2]_$F[3]";
    #my $name = "s_$line";

    my $seq1  = "";
    my $qual1 = "";
    my @c1    = ();
    for (my $i = 4; $i <= $#F; $i++) {
	my $s1 = "";
	my $q1 = -100;
	my @q = split(" ", $F[$i]);
	my $p_tot = 0;

	for (my $j = 0; $j < 4; $j++) {
	    my $prob = exp(-$q[$j]); 
	    $p_tot += $prob;

	    if ($prob > $q1) {
		$q1 = $prob; $s1 = $j;
	    }
	}

	# Renormalise and add in a chance of an external event (eg PCR)
	# causing a mutation prior to sequencing.
	$q1 = (1-$pPCR)*$q1/$p_tot + $pPCR/3;
	$q1 = int(-10*log(1-$q1)/log(10) + .5);

	$q1 = 126-33 if ($q1 > 126-33);

	$qual1 .= chr(33+$q1);
	$seq1  .= $s1;
	push(@c1, $q[4]);
	$max = $q[4] if ($max < $q[4]);
    }
    
    $seq1 =~ tr/0123/ACGT/;

    my $bytes_per_code;
    if ($max < 256) {
	$bytes_per_code = 1;
    } elsif ($max < 65536) {
	$bytes_per_code = 2;
    } else {
	$bytes_per_code = 4;
    }

    # Paired end, so split.
    my $len = length($seq1) / 2;

    if ($strand eq "-") {
	output_sam(1, $name, $rname, $rg, $start, $strand,
		   substr($seq1, $len, $len), substr($qual1, $len, $len),
		   $bytes_per_code, @c1[$len..$#c1]);

	$seq1  = reverse(substr($seq1,  0, $len));
	$seq1  =~ tr/acgtACGT/tgcaTGCA/;
	$qual1 = reverse(substr($qual1, 0, $len));
	@c1    = reverse(@c1[0..$len-1]);
	output_sam(0, $name, $rname, $rg, $end-$len, $strand,
		   $seq1, $qual1, $bytes_per_code, @c1);
    } else {
	output_sam(0, $name, $rname, $rg, $start, $strand,
		   substr($seq1, 0,    $len), substr($qual1, 0,    $len),
		   $bytes_per_code, @c1[0..$len-1]);

	$seq1  = reverse(substr($seq1,  $len, $len));
	$seq1  =~ tr/acgtACGT/tgcaTGCA/;
	$qual1 = reverse(substr($qual1, $len, $len));
	@c1    = reverse(@c1[$len..$#c1]);
	output_sam(1, $name, $rname, $rg, $end-$len+1, $strand,
		   $seq1, $qual1, $bytes_per_code, @c1);
    }
}

close($fh);

