#!/usr/bin/perl -w
use strict;

# Converts SAM with EV codes to AYB's .like output format.
#
# Usage samC2like < a.sam > a.like
#

#-----------------------------------------------------------------------------
# Read a code file.
# The format comes from Richard Durbin's proposal as follows:
# 
# > 2a) An external table file would be a text file, first with a header
# > containing the following records:
# > 
# >   #EVTABLE version         identifier that this is an event table file,
# > 			   version is a string of form <major>.<minor>
# > 
# >   #NREC <integer>          number of main records, M
# > 
# >   #MAXWORD <integer>       largest number of true bases represented by one
# > 			   code (normally 2)
# > 
# >   #ALPHABET <string>       alphabet for true sequence, normally "ACGT",
# > 			   but can allow for others, e.g. U, methyl-C
# > 
# > then a set of NREC records, each of the form
# > 
# > <CODE>\t<MARKOV>\t<DATA...>
# > 
# > where CODE runs from 1 to N, MARKOV is the markov order which is
# > typically 0 for direct base-space technologies or 1 for colour space
# > technologies, and <data> is a space separated sequence of (1 + A + A*A
# > + ...+ A^MAXWORD)*(A^MARKOV) one-byte unsigned integer log likelihood
# > values in -10log_10 units, where A is the size of the alphabet.  These
# > are interpreted as:
# > 
# >     1st value is score for this event not matching anything, i.e.  the
# >     read contains an insertion at this point compared to the truth
# > 
# >     next A values are scores for this event matching each of the
# >     alphabet character, i.e. event corresponds to one base
# > 
# >     next A2 values are scores for this event matching two bases,
# >     i.e. the read is missing a true base
# > 
# >     if MAXWORD is 3 or more, then next A3 values are scores for 
# >     this event matching 3 bases, etc.
# > 
# > Impossible (unallowed) values are encoded as 255.  If MARKOV is one,
# > then there are sets of values for each alphabet character.  Hopefully
# > Markov won't be bigger than one!
#
# For testing purposes. We only support markov 0 and maxword 2.
#
# We return a reference to an array of codes. Each code is then a hash
# consisting of keys per sequence substring. Ie -, A, C, G, T, AA, AC,
# AG, ..., TC, TG, TT. We also have a hash key "markov".

# For each base type we have 3 other base types to compare against and
# a deletion event. Ie N*M*4*4 combinations.

sub parse_codes {
    my ($fn) = @_;

    open (my $fh, $fn) || die "$fn";
    my @codes;
    
    my $nrec = 255;
    my $maxword = 2;
    my @alphabet = ("A", "C", "G", "T");
    my $na = scalar(@alphabet);

    while (<$fh>) {
	my %seqs;
	$nrec     = $1           if (/^\#NREC\s+(\d+)/);
	$maxword  = $1           if (/^\#MAXWORD\s+(\d+)/);
	@alphabet = split("",$1) if (/^\#ALPHABET\s+(\S+)/);
	next                     if (/^\#/);
	next                     if (/^\s+$/);

	my @F = split(" ", $_);
	my $c = $F[0];
	$seqs{markov} = $F[1];
	for (my $i = 2; $i <= $#F; $i++) {
	    last if $F[$i] =~ /^#/;

	    my $s;
	    if ($i <= 2) {
		$s = "-";
	    } elsif ($i <= 2 + $na) {
		$s = $alphabet[$i-3];
	    } elsif ($i <= 2 + $na + $na*$na) {
		$s = $alphabet[($i-3-$na)/$na] . $alphabet[($i-3-$na) % $na];
	    } else {
		die "Too many codes: $i of $#F vs 2+$na+$na*$na";
	    }
	    #print "i=$i, s=$s, F[]=$F[$i]\n";

	    $seqs{$s} = $F[$i];
	}
	$codes[$c] = \%seqs;
    }

    close($fh) || die "$fn";

    return \@codes;
}

#-----------------------------------------------------------------------------
# Main entry

# Load code table - supports exactly 1 only.
my $ctab;
while (<>) {
    chomp();
    if (/^\@RG/) {
	my ($ev) = (/\tEV:([^\t]*)/);
	$ctab = parse_codes($ev);
	last;
    }
}

my $last_name = "";
my $line = "";
while (<>) {
    next if (/^@/);

    my @F = split("\t", $_);

    # Name
    if ($F[0] ne $last_name) {
	print "$line\n" if $line ne "";
	$last_name = $F[0];

	my ($a,$b,$c,$d) = split("_", $F[0]);
	$line = "$a\t$b\t$c\t$d";
    }

    # Likelihoods
    my $cstr = "";
    foreach (@F) {
	next unless /^EV:Z:/;
	$cstr = $_;
    }
    $cstr =~ s/^EV:Z://;

    my @cint = split(",", $cstr);
    #my @cint = map { hex($_) } ($cstr =~ /(..)/g);

    if ($F[1] & 0x10) {
	# Reverse strand
	foreach (reverse @cint) {
	    $line .= "\t$ctab->[$_]{A} $ctab->[$_]{C} $ctab->[$_]{G} $ctab->[$_]{T}";
	}
    } else {
	foreach (@cint) {
	    $line .= "\t$ctab->[$_]{A} $ctab->[$_]{C} $ctab->[$_]{G} $ctab->[$_]{T}";
	}
    }
}
